@@ -16,6 +16,6 @@ class BaseSynthesizer:
     @classmethod
     def load(cls, path):
         device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
-        model = torch.load(path)
+        model = torch.load(path, weights_only=False)
         model.set_device(device)
-        return model
+        return model
\ No newline at end of file
