@@ -1,27 +1,29 @@
-### 2021-11-16 ###
-# pos_weight in BCEWithLogitsLoss model #
-# sigma = 5 #
-# delta = 2e-6 #
+import os
+import torch
 
+from datetime import datetime
 import warnings
-
+import joblib
 import numpy as np
 import pandas as pd
-import torch
 from packaging import version
 from torch import optim
-from torch.nn import BatchNorm1d, Dropout, LeakyReLU, Linear, Module, ReLU, Sequential, functional, BCEWithLogitsLoss, utils
+from torch.nn import BatchNorm1d, Dropout, LeakyReLU, Linear, Module, ReLU, Sequential, functional, BCEWithLogitsLoss
+from tqdm import tqdm
 
-from dp_cgans.data_sampler import DataSampler
-from dp_cgans.data_transformer import DataTransformer
+from dp_cgans.synthesizers.data_sampler import DataSampler
+from dp_cgans.synthesizers.data_transformer import DataTransformer
 from dp_cgans.synthesizers.base import BaseSynthesizer
 
-import scipy.stats
-
 ######## ADDED ########
-from datetime import datetime
 from contextlib import redirect_stdout
-from dp_cgans.rdp_accountant import compute_rdp, get_privacy_spent
+from dp_cgans.functions.rdp_accountant import compute_rdp, get_privacy_spent
+
+######## ADDED - wandb ########
+from types import SimpleNamespace
+from pathlib import Path
+import copy
+import matplotlib.pyplot as plt
 
 class Discriminator(Module):
 
@@ -53,9 +55,8 @@ class Discriminator(Module):
             create_graph=True, retain_graph=True, only_inputs=True
         )[0]
 
-        gradient_penalty = ((
-            gradients.view(-1, pac * real_data.size(1)).norm(2, dim=1) - 1
-        ) ** 2).mean() * lambda_
+        gradients_view = gradients.view(-1, pac * real_data.size(1)).norm(2, dim=1) - 1
+        gradient_penalty = ((gradients_view) ** 2).mean() * lambda_
 
         return gradient_penalty
 
@@ -140,12 +141,19 @@ class DPCGANSynthesizer(BaseSynthesizer):
             Whether to attempt to use cuda for GPU computation.
             If this is False or CUDA is not available, CPU will be used.
             Defaults to ``True``.
+        private (bool): 
+            Whether to use differential privacy
+        wandb_config (dict):
+            whether to use weights and bias tool to monitor the training
+        conditional_columns (float):
+            a matrix of embeddings
     """
 
     def __init__(self, embedding_dim=128, generator_dim=(256, 256), discriminator_dim=(256, 256),
                  generator_lr=2e-4, generator_decay=1e-6, discriminator_lr=2e-4,
                  discriminator_decay=1e-6, batch_size=500, discriminator_steps=1,
-                 log_frequency=True, verbose=False, epochs=300, pac=10, cuda=True, private=False, conditional_columns=None):
+                 log_frequency=True, verbose=False, epochs=300, pac=10, cuda=True, private=False,
+                 wandb=False, conditional_columns=None):
 
         assert batch_size % 2 == 0
 
@@ -167,6 +175,8 @@ class DPCGANSynthesizer(BaseSynthesizer):
 
         self.private = private
         self.conditional_columns = conditional_columns
+        self.wandb = wandb
+
 
         if not cuda or not torch.cuda.is_available():
             device = 'cpu'
@@ -182,6 +192,7 @@ class DPCGANSynthesizer(BaseSynthesizer):
         self._generator = None
         self._discriminator = None
 
+
     @staticmethod
     def _gumbel_softmax(logits, tau=1, hard=False, eps=1e-10, dim=-1):
         """Deals with the instability of the gumbel_softmax for older versions of torch.
@@ -227,41 +238,18 @@ class DPCGANSynthesizer(BaseSynthesizer):
                     data_t.append(transformed)
                     st = ed
                 else:
-                    assert 0
+                    raise ValueError(f'Unexpected activation function {span_info.activation_fn}.')
 
         return torch.cat(data_t, dim=1)
 
-    # def _cond_loss(self, data, c, m):
-    #     """Compute the cross entropy loss on the fixed discrete column."""
-    #     loss = []
-    #     st = 0
-    #     st_c = 0
-    #     for column_info in self._transformer.output_info_list:
-    #         for span_info in column_info:
-    #             if len(column_info) != 1 or span_info.activation_fn != "softmax":
-    #                 # not discrete column
-    #                 st += span_info.dim
-    #             else:
-    #                 ed = st + span_info.dim
-    #                 ed_c = st_c + span_info.dim
-    #                 tmp = functional.cross_entropy(
-    #                     data[:, st:ed],
-    #                     torch.argmax(c[:, st_c:ed_c], dim=1),
-    #                     reduction='none'
-    #                 )
-    #                 loss.append(tmp)
-    #                 st = ed
-    #                 st_c = ed_c
-
-    #     loss = torch.stack(loss, dim=1)
-    #     # print((loss * m).sum() / data.size()[0])
-    #     return (loss * m).sum() / data.size()[0]
 
     def _cond_loss_pair(self, data, c_pair, m_pair):
         
-        # m_pair = m_pair.detach().numpy()
         output_info_all_columns = self._transformer.output_info_list
-        loss = torch.zeros((len(data)*int((m_pair.size()[1]*(m_pair.size()[1]-1))/2),m_pair.size()[1]))
+        loss = torch.zeros(
+            (len(data)*int((m_pair.size()[1]*(m_pair.size()[1]-1))/2),m_pair.size()[1]),
+            device=self._device
+            )
         st_primary = 0
         st_primary_c = 0
         cnt = 0
@@ -294,16 +282,7 @@ class DPCGANSynthesizer(BaseSynthesizer):
                                 real_data_labels = torch.cat([data[:,st_primary:ed_primary], data[:,st_secondary:ed_secondary]], dim=1)
                                 class_counts = real_data_labels.sum(axis=0)
 
-                                
-                                # pos_weights = torch.ones_like(class_counts)
-                                # neg_counts = [len(data)-pos_count for pos_count in class_counts]
-                                # for cdx, (pos_count, neg_count) in enumerate(zip(class_counts,  neg_counts)):
-                                #     pos_weights[cdx] = neg_count / (pos_count + 1e-5)
-                                
-                                # torch_pos_weights = torch.as_tensor(pos_weights, dtype=torch.float)
-                                # print(pos_weights)
-
-                                criterion = BCEWithLogitsLoss(reduction='none')#, pos_weight=pos_weights)
+                                criterion = BCEWithLogitsLoss(reduction='none').to(self._device)#, pos_weight=pos_weights)
                                 calculate_loss = criterion(
                                     torch.cat([data[:,st_primary:ed_primary], data[:,st_secondary:ed_secondary]], dim=1),
                                     torch.cat([c_pair[:,st_primary_c:ed_primary_c], c_pair[:,st_secondary_c:ed_secondary_c]],dim=1)
@@ -321,8 +300,9 @@ class DPCGANSynthesizer(BaseSynthesizer):
                     cnt_primary += 1
                     st_primary = ed_primary
                     st_primary_c = ed_primary_c
-        # print(len(loss))
-        return loss.sum() / len(loss)
+        
+        result = loss.sum() / len(loss)
+        return result
 
 
     def _validate_discrete_columns(self, train_data, discrete_columns):
@@ -354,8 +334,7 @@ class DPCGANSynthesizer(BaseSynthesizer):
     ############ Tensorflow Privacy Measurement ##############
 
     def fit(self, train_data, discrete_columns=tuple(), epochs=None):
-        """Fit the CTGAN Synthesizer models to the training data.
-
+        """
         Args:
             train_data (numpy.ndarray or pandas.DataFrame):
                 Training Data. It must be a 2-dimensional numpy array or a pandas.DataFrame.
@@ -366,12 +345,29 @@ class DPCGANSynthesizer(BaseSynthesizer):
                 a ``pandas.DataFrame``, this list should contain the column names.
         """
 
+        if self.wandb == True:
+            config = SimpleNamespace(
+                epochs=epochs, # number of training epochs
+                batch_size=self._batch_size, # the size of each batch
+                log_frequency=self._log_frequency,
+                verbose=self._verbose,
+                generator_dim=self._generator_dim,
+                discriminator_dim=self._discriminator_dim,
+                generator_lr=self._generator_lr,
+                discriminator_lr=self._discriminator_lr,
+                discriminator_steps=self._discriminator_steps, 
+                private=self.private
+                
+            )
 
-        # if self.conditional_columns != None:
-        #     if set(self.conditional_columns) <= set(discrete_columns):
-        #         discrete_columns = self.conditional_columns
-        #     else:
-        #         raise NotImplementedError("Conditional columns are not in the valid columns.",discrete_columns)
+            # wandb_config = wandb.init(
+            #     project="dp_cgans_training_monitor",
+            #     anonymous="allow",
+            #     config=config
+            # )
+ 
+        real_data_columns = [col for col in train_data.columns if '.value' in col]
+        real_data = copy.deepcopy(train_data[real_data_columns])
 
 
         self._validate_discrete_columns(train_data, discrete_columns)
@@ -386,16 +382,30 @@ class DPCGANSynthesizer(BaseSynthesizer):
             )
 
         self._transformer = DataTransformer()
-        self._transformer.fit(train_data, discrete_columns)
+
+        print("Start transforming data...")
+        
+        if os.path.exists(os.getcwd()+'/fitted_transformer.pkl'):
+            print("Loading fitted transformer...")
+            self._transformer = joblib.load(os.getcwd()+'/fitted_transformer.pkl')
+        else:
+            print("Start fitting transformer ...")
+            self._transformer.fit(train_data, discrete_columns)
+            self._transformer.fit(train_data, discrete_columns)
+            joblib.dump(self._transformer, os.getcwd()+'/fitted_transformer.pkl')
+            print("Saving fitted transformer...")
+        
+        print("Finish transforming data / loading transformed data...")
 
         train_data = self._transformer.transform(train_data)
-
+        
         self._data_sampler = DataSampler(
             train_data,
             self._transformer.output_info_list,
             self._log_frequency)
 
         data_dim = self._transformer.output_dimensions
+        
 
         self._generator = Generator(
             self._embedding_dim + self._data_sampler.dim_cond_vec(), # number of categories in the whole dataset.
@@ -422,194 +432,199 @@ class DPCGANSynthesizer(BaseSynthesizer):
         mean = torch.zeros(self._batch_size, self._embedding_dim, device=self._device)
         std = mean + 1
 
-        steps_per_epoch = max(len(train_data) // self._batch_size, 1)
-        ######## ADDED ########
-        with open('loss_output_%s.txt'%str(epochs), 'w') as f:
-            with redirect_stdout(f):
-                ######## ADDED ########
-                for i in range(epochs):
-                    for id_ in range(steps_per_epoch):
+        self.loss_values = pd.DataFrame(columns=['Epoch', 'Generator Loss', 'Distriminator Loss'])
+        
+        epoch_iterator = tqdm(range(epochs), disable=(not self._verbose))
+        if self._verbose:
+            description = 'Gen. ({gen:.2f}) | Discrim. ({dis:.2f})'
+            epoch_iterator.set_description(description.format(gen=0, dis=0))
 
-                        for n in range(self._discriminator_steps):
-                            fakez = torch.normal(mean=mean, std=std)
-
-                            # condvec = self._data_sampler.sample_condvec(self._batch_size)
-                            
-                            condvec_pair = self._data_sampler.sample_condvec_pair(self._batch_size)
-                            c_pair_1, m_pair_1, col_pair_1, opt_pair_1 = condvec_pair
-
-                            # if condvec is None:
-                            #     c1, m1, col, opt = None, None, None, None
-                            #     real = self._data_sampler.sample_data(self._batch_size, col, opt)
-                            # else:
-                            #     c1, m1, col, opt = condvec
-                            #     c1 = torch.from_numpy(c1).to(self._device)
-                            #     m1 = torch.from_numpy(m1).to(self._device)
-                            #     fakez = torch.cat([fakez, c1], dim=1)
-
-                            #     perm = np.arange(self._batch_size)
-                            #     np.random.shuffle(perm)
-                            #     real = self._data_sampler.sample_data(
-                            #         self._batch_size, col[perm], opt[perm])
-                            #     c2 = c1[perm]
-
-                            if condvec_pair is None:
-                                c_pair_1, m_pair_1, col_pair_1, opt_pair_1 = None, None, None, None
-                                real = self._data_sampler.sample_data_pair(self._batch_size, col_pair_1, opt_pair_1)
-                            else:
-                                c_pair_1, m_pair_1, col_pair_1, opt_pair_1 = condvec_pair
-                                c_pair_1 = torch.from_numpy(c_pair_1).to(self._device)
-                                m_pair_1 = torch.from_numpy(m_pair_1).to(self._device)
-                                fakez = torch.cat([fakez, c_pair_1], dim=1)
-
-                                perm = np.arange(self._batch_size)
-                                np.random.shuffle(perm)
             
-                                real = self._data_sampler.sample_data_pair(self._batch_size, col_pair_1[perm], opt_pair_1[perm])
-                                c_pair_2 = c_pair_1[perm]
+        steps_per_epoch = max(len(train_data) // self._batch_size, 1)
 
+        print(f"{datetime.now()}", 'epoch: ', epoch_iterator)
+        filename = f'loss_output_{datetime.now().strftime("%Y%m%d_%H%M%S")}.txt'
+        with open(filename, 'w') as loss_file:
+            loss_file.write("Epoch,Generator_Loss,Discriminator_Loss,Timestamp\n")
+            for i in epoch_iterator:
+                if i == 0 and self._device == 'cuda':
+                    torch.cuda.init()
+                    torch.cuda.synchronize()
+                    print("CUDA initialized:", torch.cuda.get_device_name(0))
+                for id_ in range(steps_per_epoch):
+                    print(f"{datetime.now()}", "steps_per_epoch", id_)
+                    for n in range(self._discriminator_steps):
 
-                            fake = self._generator(fakez) # categories (unique value count) + continuous (1+n_components)
-                            fakeact = self._apply_activate(fake)
-
-                            real = torch.from_numpy(real.astype('float32')).to(self._device)
-                            
-                            # if c1 is not None:
-                                # fake_cat = torch.cat([fakeact, c1], dim=1)
-                                # real_cat = torch.cat([real, c2], dim=1)
-                            if col_pair_1 is not None:
-                                fake_cat = torch.cat([fakeact, c_pair_1], dim=1)
-                                real_cat = torch.cat([real, c_pair_2], dim=1)
-                            else:
-                                real_cat = real
-                                fake_cat = fake
-
-                            y_fake = self._discriminator(fake_cat)
-                            y_real = self._discriminator(real_cat)
-
-                            loss_d = -(torch.mean(y_real) - torch.mean(y_fake))
-
-     
-                            #### DP ####
-                            if self.private:
-                                sigma = 1
-                                weight_clip = 0.01 
-
-                                if sigma is not None:
-                                    for parameter in self._discriminator.parameters():
-                                        parameter.register_hook(
-                                            lambda grad: grad.cuda() + (1 / self._batch_size) * sigma
-                                            * torch.randn(parameter.shape).cuda()
-                                        )
-                            #### DP ####
-
-                            pen = self._discriminator.calc_gradient_penalty(
-                                real_cat, fake_cat, self._device, self.pac)
-
-                            optimizerD.zero_grad()
-                            pen.backward(retain_graph=True) # https://machinelearningmastery.com/how-to-implement-wasserstein-loss-for-generative-adversarial-networks/ 
-                            loss_d.backward()
-                            optimizerD.step()
-
-                            if self.private:
-                                #### DP ####
-                                # Weight clipping for privacy guarantee
-                                for param in self._discriminator.parameters():
-                                    param.data.clamp_(-weight_clip, weight_clip)
-                                #### DP ####
-
-                        fakez = torch.normal(mean=mean, std=std)
+                        fakez = torch.normal(mean=mean, std=std).to(self._device)
+                        
                         condvec_pair = self._data_sampler.sample_condvec_pair(self._batch_size)
-
-                        # if condvec is None:
-                        #     c1, m1, col, opt = None, None, None, None
-                        # else:
-                        #     c1, m1, col, opt = condvec
-          
-                        #     c1 = torch.from_numpy(c1).to(self._device)
-                        #     m1 = torch.from_numpy(m1).to(self._device)
-                        #     fakez = torch.cat([fakez, c1], dim=1)
+                        c_pair_1, m_pair_1, col_pair_1, opt_pair_1 = condvec_pair
 
 
                         if condvec_pair is None:
                             c_pair_1, m_pair_1, col_pair_1, opt_pair_1 = None, None, None, None
+                            real = self._data_sampler.sample_data_pair(self._batch_size, col_pair_1, opt_pair_1)
                         else:
                             c_pair_1, m_pair_1, col_pair_1, opt_pair_1 = condvec_pair
-          
                             c_pair_1 = torch.from_numpy(c_pair_1).to(self._device)
                             m_pair_1 = torch.from_numpy(m_pair_1).to(self._device)
                             fakez = torch.cat([fakez, c_pair_1], dim=1)
-            
 
-                        fake = self._generator(fakez)
+                            perm = np.arange(self._batch_size)
+                            np.random.shuffle(perm)
+        
+                            real = self._data_sampler.sample_data_pair(self._batch_size, col_pair_1[perm], opt_pair_1[perm])
+                            c_pair_2 = c_pair_1[perm]
+
+
+                        real = torch.from_numpy(real.astype('float32')).to(self._device)
+                        
+                        fake = self._generator(fakez) # categories (unique value count) + continuous (1+n_components)
                         fakeact = self._apply_activate(fake)
 
-                        # if c1 is not None:
-                        #     y_fake = discriminator(torch.cat([fakeact, c1], dim=1))
-                        # else:
-                        #     y_fake = discriminator(fakeact)
-
-                        # if condvec is None:
-                        #     cross_entropy = 0
-                        # else:
-                        #     cross_entropy = self._cond_loss(fake, c1, m1)
-
-                        # loss_g = -torch.mean(y_fake) + cross_entropy# + rules_penalty
-
-                        if c_pair_1 is not None:
-                            y_fake = self._discriminator(torch.cat([fakeact, c_pair_1], dim=1))
+                        
+                        
+                        if col_pair_1 is not None:
+                            fake_cat = torch.cat([fakeact, c_pair_1], dim=1)
+                            real_cat = torch.cat([real, c_pair_2], dim=1)
                         else:
-                            y_fake = self._discriminator(fakeact)
-
-                        if condvec_pair is None:
-                            cross_entropy_pair = 0
-                        else:
-                            cross_entropy_pair = self._cond_loss_pair(fake, c_pair_1, m_pair_1)
-
-
-                        loss_g = -torch.mean(y_fake) + cross_entropy_pair # + rules_penalty
+                            real_cat = real
+                            fake_cat = fake
                         
 
-                        optimizerG.zero_grad()
-                        loss_g.backward()
-                        optimizerG.step()
-                    
-                    if self._verbose:
-                        ######## ADDED ########
-                        now = datetime.now()
-                        current_time = now.strftime("%H:%M:%S")
-
-                        # Calculate the current privacy cost using the accountant
-                        # https://github.com/BorealisAI/private-data-generation/blob/master/models/dp_wgan.py
-                        # https://github.com/tensorflow/privacy/tree/master/tutorials/walkthrough
-
-                        print(current_time, f"Epoch {i+1}, Loss G: {loss_g.detach().cpu(): .4f},"
-                            f"Loss D: {loss_d.detach().cpu(): .4f}", flush=True)
+                        y_fake = self._discriminator(fake_cat)
+                        y_real = self._discriminator(real_cat)
+                        loss_d = -(torch.mean(y_real) - torch.mean(y_fake))
 
 
+                        #### DP ####
                         if self.private:
-                            orders = [1 + x / 10. for x in range(1, 100)]
-                            sampling_probability = self._batch_size/len(train_data)
-                            delta = 2e-6
-                            rdp = compute_rdp(q=sampling_probability,
-                                                noise_multiplier=sigma,
-                                                steps=i * steps_per_epoch,
-                                                orders=orders)
-                            epsilon, _, opt_order = get_privacy_spent(orders, rdp, target_delta=delta) # target_delta=1e-5
+                            sigma = 1
+                            weight_clip = 0.01 
 
-                            print('differential privacy with eps = {:.3g} and delta = {}.'.format(
-                                epsilon, delta))
-                            print('The optimal RDP order is {}.'.format(opt_order))
+                            if sigma is not None:
+                                for parameter in self._discriminator.parameters():
+                                    parameter.register_hook(
+                                        lambda grad: grad.cuda() + (1 / self._batch_size) * sigma
+                                        * torch.randn(parameter.shape).cuda()
+                                    )
+                        #### DP ####
+                        pen = self._discriminator.calc_gradient_penalty(
+                            real_cat, fake_cat, self._device, self.pac)
 
-                            if opt_order == max(orders) or opt_order == min(orders):
-                                print('The privacy estimate is likely to be improved by expanding '
-                                    'the set of orders.')
-                        else:
-                            epsilon = np.nan
+                        optimizerD.zero_grad()
+                        # https://machinelearningmastery.com/how-to-implement-wasserstein-loss-for-generative-adversarial-networks/ 
+                        pen.backward(retain_graph=True) 
+                        loss_d.backward()
+                        optimizerD.step()
+                        if self.private:
+                            #### DP ####
+                            # Weight clipping for privacy guarantee
+                            for param in self._discriminator.parameters():
+                                param.data.clamp_(-weight_clip, weight_clip)
+                            #### DP ####
 
-                        
+                    fakez = torch.normal(mean=mean, std=std)
+                    condvec_pair = self._data_sampler.sample_condvec_pair(self._batch_size)
+
+                    if condvec_pair is None:
+                        c_pair_1, m_pair_1, col_pair_1, opt_pair_1 = None, None, None, None
+                    else:
+                        c_pair_1, m_pair_1, col_pair_1, opt_pair_1 = condvec_pair
+        
+                        c_pair_1 = torch.from_numpy(c_pair_1).to(self._device)
+                        m_pair_1 = torch.from_numpy(m_pair_1).to(self._device)
+                        fakez = torch.cat([fakez, c_pair_1], dim=1)
+        
+
+                    fake = self._generator(fakez)
+                    fakeact = self._apply_activate(fake)
+
+                    if c_pair_1 is not None:
+                        y_fake = self._discriminator(torch.cat([fakeact, c_pair_1], dim=1))
+                    else:
+                        y_fake = self._discriminator(fakeact)
+
+                    if condvec_pair is None:
+                        cross_entropy_pair = 0
+                    else:
+                        cross_entropy_pair = self._cond_loss_pair(fake, c_pair_1, m_pair_1)
+
+                    # loss_g_pure =  -torch.mean(y_fake)
+                    loss_g = -torch.mean(y_fake) + cross_entropy_pair # + rules_penalty
+
+                    # Before backward
+                    if self._device == 'cuda':
+                        print(f"GPU allocated before backward: {torch.cuda.memory_allocated() / 1024**2:.1f} MB")
+                        print(f"GPU reserved before backward: {torch.cuda.memory_reserved() / 1024**2:.1f} MB")
+                        torch.cuda.reset_peak_memory_stats()
+
+                    optimizerG.zero_grad(set_to_none=False)
+                    loss_g.backward()
+                    optimizerG.step()
+
+                    if self._device == 'cuda':
+                        print(f"GPU allocated after backward: {torch.cuda.memory_allocated() / 1024**2:.1f} MB")
+                        print(f"GPU peak memory: {torch.cuda.max_memory_allocated() / 1024**2:.1f} MB")
+                        # print("If peak memory is small, backward computation is on CPU!")
+
+                generator_loss = loss_g.detach().cpu().item()
+                discriminator_loss = loss_d.detach().cpu().item()
+                
+                timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
+                loss_file.write(f"{i},{generator_loss:.3f},{discriminator_loss:.3f},{timestamp}\n")
+                loss_file.flush()  # Write immediately to disk
+
+                print(f"{datetime.now()}", 'Epoch', [i], '; Generator Loss', [generator_loss],'; Discriminator Loss',[discriminator_loss])
+
+
+                if self._verbose:
+                    epoch_iterator.set_description(
+                        description.format(gen=generator_loss, dis=discriminator_loss)
+                    )
+
+                if self.wandb == True :
+                    ## Add WB logs
+                    metrics = {
+                        # "train/loss_g_pure": loss_g_pure.detach().cpu(),
+                        "train/loss_g": loss_g.detach().cpu(),
+                        "train/loss_d": loss_d.detach().cpu(),
+                        "train/epoch": i + 1,
+                        #"train/example_ct": len(loss_g)
+                    }
+                    # wandb.log(metrics)
+
+                    SAVE_DIR = Path('./data/weights/')
+                    SAVE_DIR.mkdir(exist_ok=True, parents=True)
+
+
+                if self.private:
+                    orders = [1 + x / 10. for x in range(1, 100)]
+                    sampling_probability = self._batch_size/len(train_data)
+                    delta = 2e-6
+                    rdp = compute_rdp(q=sampling_probability,
+                                        noise_multiplier=sigma,
+                                        steps=i * steps_per_epoch,
+                                        orders=orders)
+                    epsilon, _, opt_order = get_privacy_spent(orders, rdp, target_delta=delta) # target_delta=1e-5
+
+                    print('differential privacy with eps = {:.3g} and delta = {}.'.format(
+                        epsilon, delta))
+                    print('The optimal RDP order is {}.'.format(opt_order))
+
+                    if opt_order == max(orders) or opt_order == min(orders):
+                        print('The privacy estimate is likely to be improved by expanding '
+                            'the set of orders.')
+                else:
+                    epsilon = np.nan
+
+                            
                     ######## ADDED ########
-                    
+                if self.wandb == True:
+                    wandb.finish()
+
+            
+    
 
     def sample(self, n, condition_column=None, condition_value=None):
         """Sample data similar to the training data.
